#import "ScenarioSource.h"
#import <Foundation/NSString.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSFileManager.h>
#import <Foundation/NSNotification.h>
#import <Foundation/NSBundle.h>	/* LocalizedString */
#import <AppKit/NSImage.h>
#import "ScenarioCtrl.h"
#import "InspectorCtrl.h"
#import "SettingDictionary.h"
#import "TextImageCreator.h"
#import "commtypes.h"
#import "dickeys.h"

#define  SkipMark	@"--"
#define  ErrMark	@"EE"

enum {	/* Error Code of "checkFile:inDir:" */
	Code_OK, Code_NotExist, Code_CannotRead, Code_IsDir
};

@implementation ScenarioSource

- (id)init
{
	[super init];
	records = [[NSMutableArray alloc] initWithCapacity:3];
	controller = nil;
	return self;
}

- (void)setController:(id)ctrl
{
	controller = ctrl;
}

- (unsigned)count
{
	return [records count];
}

- (void)castNotification
{
	[[NSNotificationCenter defaultCenter]
		postNotificationName:ScenarioDidChange object:self];
}

/* Local Method */
- (int)checkFile:(NSString *)file inDir:(NSString *)dir
{
	BOOL	isDir;
	NSString *path = [dir stringByAppendingPathComponent:file];
	NSFileManager *manager = [NSFileManager defaultManager];
	if (![manager fileExistsAtPath:path isDirectory:&isDir])
		return Code_NotExist;
	if (isDir && ![TextImageCreator isWrappedFile:path])
		return Code_IsDir;
	if (![manager isReadableFileAtPath:path])
		return Code_CannotRead;
	return Code_OK;
}

/* Local Method */
- (void)addFilename:(NSString *)name withCode:(int)code asIndex:(int)num
{
	NSString *msg = nil;
	id r = [[NSMutableDictionary alloc] initWithCapacity:1];
	[r setObject:name forKey:KFilename];
	switch (code) {
	case Code_NotExist:
		msg = NSLocalizedString(@"No such file", Err_NoFile);
		break;
	case Code_CannotRead:
		msg = NSLocalizedString(@"Can not read", Err_Read);
		break;
	case Code_IsDir:
		msg = NSLocalizedString(@"It is a directory", Err_Directory);
		break;
	case Code_OK:
	default:
		break;
	}
	if (msg) {
		[r setObject:msg forKey:KMemo];
		if (code == Code_NotExist || code == Code_CannotRead)
			[r setObject:CanNotRead forKey:KSkip];
	}
	[records insertObject:r atIndex:num];
}

- (int)addFiles:(NSArray *)flist inDir:(NSString *)dir fromIndex:(int)row
{
	int	n, i, code;
	NSString *fn;

	if ((n = [flist count]) == 0)
		return 0;
	if (n > 1)
		flist = [flist sortedArrayUsingSelector:
					@selector(caseInsensitiveCompare:)];
	for (i = 0; i < n; i++) {
		fn = [flist objectAtIndex:i];
		code = [self checkFile:fn inDir:dir];
		[self addFilename:fn withCode:code asIndex:row++];
	}
	[self castNotification];
	return n;
}

- (void)removeAtIndex:(int)num
{
	if ([records count] <= num)
		return;
	[records removeObjectAtIndex:num];
	[self castNotification];
}

- (int)checkFileExistence
{
	int	i, num, count, code;
	id	dic;
	NSString *dir, *skip;

	if ((num = [records count]) <= 0)
		return -1;
	dir = [[controller filename] stringByDeletingLastPathComponent];
	count = 0;
	for (i = 0; i < num; i++) {
		dic = [records objectAtIndex:i];
		skip = [dic objectForKey:KSkip];
		if (skip && [skip isEqualToString:SkipTrue])
			continue;
		code = [self checkFile:[dic objectForKey:KFilename] inDir:dir];
		if (code == Code_NotExist || code == Code_CannotRead) {
			[dic setObject:CanNotRead forKey:KSkip];
			count++;
		}else if (skip /* && [skip isEqualToString:CanNotRead] */)
			[dic removeObjectForKey:KSkip];
	}
	if (count > 0)
		[self castNotification];
		/* also "reloadData" from the table is needed */
	return count;
}

- (void)doubleClicked:(id)sender
{
	int	row, code;
	id	col, rec, skip;
	NSString *ids, *dir, *fn;

	if ((row = [sender clickedRow]) < 0)
		return;		/* column header is selected */
	col = [[sender tableColumns] objectAtIndex:[sender clickedColumn]];
	ids = [col identifier];
	if ([ids isEqualToString:KNumber]) {
		rec = [records objectAtIndex:row];
		if ((skip = [rec objectForKey:KSkip]) == nil)
			[rec setObject:SkipTrue forKey:KSkip];
		else {
			dir = [[controller filename] stringByDeletingLastPathComponent];
			fn = [rec objectForKey:KFilename];
			code = [self checkFile:fn inDir:dir];
			if (code == Code_NotExist || code == Code_CannotRead) {
				NSBeep();
				if ([skip isEqualToString:CanNotRead])
				    return;
					/* No need to reloadData */
				else
				    [rec setObject:CanNotRead forKey:KSkip];
			}else
				[rec removeObjectForKey:KSkip];
		}
		[sender reloadData];
		[self castNotification];
	}else if ([ids isEqualToString:KFilename]) {
		rec = [records objectAtIndex:row];
		[controller openFileByWS:[rec objectForKey:KFilename]];
	}
}

- (NSArray *)fileListUsed
{
	int	i, n;
	id	r, list;

	n = [records count];
	list = [NSMutableArray arrayWithCapacity: n];
	for (i = 0; i < n; i++) {
		r = [[records objectAtIndex: i] objectForKey:KFilename];
		if (r)
			[list addObject: r];
	}
	return list;
}

- (void)insertDirectoryElement:(NSString *)dir
	/* used when "Save As" operation is done */
{
	int	i, n;
	id	rec;
	NSString *newf;

	if (dir == nil || [dir length] == 0)
		return;
	for (i = 0, n = [records count]; i < n; i++) {
		rec = [records objectAtIndex:i];
		newf = [dir stringByAppendingPathComponent:
				[rec objectForKey:KFilename]];
		[rec setObject:newf forKey:KFilename];
	}
	[self castNotification];
}

- (BOOL)writeToFile:(NSString *)fpath with:(id)setting
{
	id dic = [NSMutableDictionary dictionaryWithCapacity:2];
	[dic setObject:setting forKey:DSetting];
	[dic setObject:records forKey:DScenario];
	return [dic writeToFile:fpath atomically:NO];
}

- (id)readFromFile:(NSString *)fpath
	/* This method returns object(autorelease) of SettingDictionary */
{
	id	dic, setting, arr;
	int	n, i;

	if ((dic = [NSDictionary dictionaryWithContentsOfFile:fpath]) == nil)
		return nil;
	if ((arr = [dic objectForKey:DScenario]) == nil)
		return nil;
	/* arr is a concrete array, not mutable. */
	/* records = [arr mutableCopy] <-- each element is still concrete */
	n = [arr count];
	[records release];
	records = [[NSMutableArray alloc] initWithCapacity:n];
	for (i = 0; i < n; i++)
		[records addObject:
			[[NSMutableDictionary alloc]
				initWithDictionary:[arr objectAtIndex:i]]];
	setting = [[SettingDictionary alloc]
                initWithDictionary:[dic objectForKey:DSetting]];
	[setting autorelease];
	return setting;
}

/* Used from Inspector & SlideShow */
- (NSString *)optionStringAtIndex:(int)rowIndex
{
	return [[records objectAtIndex:rowIndex] objectForKey:KOption];
}

/* Used from Inspector & SlideShow */
- (NSString *)filenameAtIndex:(int)rowIndex
{
	id rec = [records objectAtIndex:rowIndex];
	if ([rec objectForKey:KSkip] != nil)
		return nil;
	return [rec objectForKey:KFilename];
}

/* Used from Inspector & SlideShow */
- (NSString *)titleAtIndex:(int)rowIndex
{
	return [[records objectAtIndex:rowIndex] objectForKey:KTitle];
}

/* Used from Inspector */
- (BOOL)setOptionString:(NSString *)obj atIndex:(int)rowIndex
{
	id theRecord;
	if ((theRecord = [records objectAtIndex:rowIndex]) == nil)
		return NO;
	if (![[theRecord objectForKey:KOption] isEqual:obj]) {
		[theRecord setObject:obj forKey:KOption];
		[self castNotification];
		return YES;
	}
	return NO;
}

/* NSTableDataSource */
- (int)numberOfRowsInTableView:(NSTableView *)tableView
{
	return [records count];
}

- (id)tableView:(NSTableView *)aTableView
    objectValueForTableColumn:(NSTableColumn *)aTableColumn
    row:(int)rowIndex
{
	id theRecord, theValue;
	NSString *ids, *skip;

	// NSParameterAssert(rowIndex >= 0 && rowIndex < [records count]);
	theRecord = [records objectAtIndex:rowIndex];
	ids = [aTableColumn identifier];
	if ([ids isEqualToString:KNumber]) {
		skip = [theRecord objectForKey:KSkip];
		if (skip != nil)
			return [skip isEqualToString:SkipTrue] ? SkipMark : ErrMark;
		return [NSString stringWithFormat:@"%d", rowIndex+1];
	}

	theValue = [theRecord objectForKey:[aTableColumn identifier]];
	return theValue;
}


- (void)tableView:(NSTableView *)aTableView setObjectValue:anObject
	forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
	id theRecord, theObj, key;

	// NSParameterAssert(rowIndex >= 0 && rowIndex < [records count]);
	key = [aTableColumn identifier];
	theRecord = [records objectAtIndex:rowIndex];
	theObj = [theRecord objectForKey:key];
	if (anObject == nil || [anObject isEqual:@""]) {
		if (theObj == nil)
			return;
		[theRecord removeObjectForKey: key];
//	Note: [anObject isEqual:@""] is almost equal to
//	 ([anObject isKindOfClass:[NSString class]] && [anObject length] == 0)
	}else if ([theObj isEqual:anObject])
		return;
	else
		[theRecord setObject:anObject forKey:key];
	[self castNotification];
	if ([key isEqualToString:KOption])
		[[InspectorCtrl defaultController] setCurrentTable:aTableView];
}

- (BOOL)tableView:(NSTableView *)aTableView shouldEditTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
	NSString *ids = [aTableColumn identifier];
	if ([ids isEqualToString:KFilename])
		return NO;
	if ([ids isEqualToString:KNumber])
		return NO;
	return YES;
}

/*
 * SDMovingRowsProtocol
 */

- (unsigned int)dragReorderingMask:(int)forColumn
{
	return DRAG_ALWAYS;
}

// Delegate called after the reordering of cells, you must reorder your data.
// Returning YES will cause the table to be reloaded.
- (BOOL)tableView:(NSTableView *)tv didDepositRow:(int)rowToMove at:(int)newPosition
{
	int	n;
	id	elm;

	n = [records count];
	if (rowToMove >= n || newPosition >= n)
		return NO;
	if (rowToMove == newPosition)
		return YES;
	elm = [records objectAtIndex:rowToMove];
	if (rowToMove > newPosition) {
		[records insertObject:elm atIndex:newPosition];
		[records removeObjectAtIndex:(rowToMove + 1)];
	}else {
		[records insertObject:elm atIndex:(newPosition + 1)];
		[records removeObjectAtIndex: rowToMove];
	}
	[self castNotification];
	return YES;
}

// Notifies the delegate of impending movement
- (void)willMoveRow:(int)row
{
}

// When we move an image around in the reordering process
// we'll use this method to get the correct image:
// - (NSImage *)imageForRow:(int)row
//{
//	NSRect rect = [theView rectOfRow:row];
//	return [[[NSImage alloc] initWithData:
//			[theView dataWithEPSInsideRect:rect]] autorelease];
//}

// This gives you a chance to decline to drop particular rows on other
// particular row. Return YES if you don't care
- (BOOL) tableView:(NSTableView *)tableView draggingRow:(int)draggedRow overRow:(int) targetRow
{
	return YES;
}

@end
