#import "SlideShowCtrl.h"
#import <Foundation/NSArray.h>
#import <Foundation/NSLock.h>
#import <Foundation/NSThread.h>
#import <Foundation/NSAutoreleasePool.h>
#import <AppKit/NSColor.h>
#import <AppKit/NSTableView.h>
#import "FullScreenView.h"
#import "FullScreenWindow.h"
#import "ScenarioSource.h"
#import "TextImageCreator.h"
#import "PrefCtrl.h"

@implementation SlideShowCtrl

- (id)init
{
	[super init];
	scrRect.origin.x = scrRect.origin.y = 0;
	scrRect.size = [[PrefCtrl defaultPref] screenSize];
	backView = [[FullScreenView alloc] initWithFrame: scrRect];
	backWin = [[FullScreenWindow alloc] initWithView: backView];
	[backWin setReleasedWhenClosed:YES];
	[backWin makeFirstResponder: backView];
	qarray = NULL;
	qLock = [[NSLock alloc] init];
	quitTarget = self;
	quitSel = @selector(release);
	return self;
}

- (id)initWithTable:(id)tbl andSetting:(struct setinfo *)info
{
	// int	i;

	[self init];
	table = tbl;
	setting = info;
	[backView setController:self];
	[backView setBackgroundColor:setting->bgcolor];
	return self;
}

- (void)setOHPPath:(NSString *)path
{
	[ohpPath release];
	[(ohpPath = path) retain];
}

- (void)setQuitAction:(SEL)sel to:(id)target
{
	quitTarget = target;
	quitSel = sel;
}

/* Local Method */
- (struct qcell *)allocQcell
{
	struct qcell *cell = (struct qcell *)malloc(sizeof(struct qcell));
	cell->img = cell->title = nil;
	return cell;
}

/* Local Method */
- (void)freeQcell:(struct qcell *)cell
{
	/* Note that this method is not guarded by LOCK */
	[cell->img release];
	[cell->title release];
	free((void *)cell);
}

/* Local Method */
- (void)freeQarray
{
	if (qarray) {
		int i;
		[qLock lock];
		for (i = 0; i < sheetNum; i++)
			if (qarray[i])
				[self freeQcell:qarray[i]];
		free((void *)qarray);
		qarray = NULL;
		[qLock unlock];
	}
}

- (void)dealloc
{
	[backView release];
	[backWin release];
	[ohpPath release];
	[self freeQarray];
	[qLock release];  /* must be release'd after freeQarray */
	[super dealloc];
}

/* Local Method */
- (NSString *)fetchFileAndOpt:(opt_t *)optp atRow:(int)row
{
	NSString *fn, *opstr;
	id	src;

	src = [table dataSource];
	if ((fn = [src filenameAtIndex:row]) == nil)
		return nil;	/* SKIP */
	opstr = [src optionStringAtIndex:row];
	if (opstr == nil) {
		opstr = @"";
		clearOption(optp);
	}else
		convStringToOption(opstr, optp);
	return fn;
}

/* Local Method */
- (void)drawRectArea:(NSSize)sz kind:(int)kind
{
	NSColor *clr;
	float transp;

	switch (kind) {
	case img_EPS:
		if (setting->epsbg == EPS_BG_none)
			break;
		if (setting->epsbg == EPS_BG_White) {
			clr = [NSColor whiteColor];
			transp = 1.0;
		}else if (setting->usetxbg /* && EPS_BG_Text */) {
			clr = setting->txbgcolor;
			transp = setting->transparent;
		}else /* setting->usetxbg == 0 && EPS_BG_Text */
			return;
		[backView drawRect:sz withColor:clr andTransp:transp];
		break;
	case img_Text:
		if (setting->usetxbg)
			[backView drawRect:sz withColor:setting->txbgcolor
				andTransp:setting->transparent];
		break;
	case img_Pict:
	default:
		break;
	}
}

/* Local Method */
- (void)showTitle:(const struct qcell *)qc imageSize:(NSSize)imgsz
{
	const float	Margin = 0.02;	/* ratio to height of screen */
	const float	MaxSpace = 0.02;
	NSSize	sz;
	NSPoint	pnt;
	float	gap, w, space = 0.0;
	NSColor	*txclr = nil, *bgclr = nil;
	float	trans = 1.0;

	if (qc->opt.notextclr == 0)
		txclr = setting->txcolor;
	sz = [qc->title size];
	gap = (scrRect.size.height - imgsz.height) / 2;
	w = scrRect.size.height * (Margin + MaxSpace) + sz.height;
	if (gap >= w)
		space = scrRect.size.height * MaxSpace;
	else
		space = gap - (scrRect.size.height * Margin + sz.height);
	if (space >= 0) {
		/* think about background color of the title */
		if (setting->usetxbg) {
			bgclr = setting->txbgcolor;
			trans = setting->transparent;
		}
	}else {
		/* think about background color of the title */
		if (txclr == nil)
			bgclr = [NSColor whiteColor];
		else {
			float bright = [[txclr colorUsingColorSpaceName:
			    NSCalibratedRGBColorSpace] brightnessComponent];
			bgclr = (bright < 0.5)
				? [NSColor whiteColor] : [NSColor blackColor];
		}
	}
	if (qc->opt.title == 1) /* Upper */
		pnt.y = imgsz.height + gap + space;
	else /* Lower */
		pnt.y = gap - (space + sz.height);
	pnt.x = (scrRect.size.width - sz.width) / 2;
	[backView drawImage:qc->title
		at:pnt withBackground:bgclr transparent:trans];
}

/* Local Method */
- (void)setTitleAtIndex:(int)index
{
	NSString *title;
	id	creat;
	struct qcell	*qc;

	qc = qarray[index];
	title = [[table dataSource] titleAtIndex:index];
	if (title == nil || [title length] == 0) {
		qc->title = nil;
		return;	/* No Title */
	}
	creat = [[TextImageCreator alloc] init];
	if (qc->opt.notextclr == 0)
		[creat setTextColor: setting->txcolor];
	[creat setFont:[setting->txfont fontName]
		size: (qc->opt.fontsize ?
			qc->opt.fontsize : setting->fontsize)];
	qc->title = [creat textImageFromString:title];
	[qc->title retain];
	[creat release];
}

/* Local Method */
- (BOOL)showSheet:(const struct qcell *)qc forward:(BOOL)forward
{
	NSSize	sz;	/* size of the image */

	if (qc->opt.backgflag) {
		[backView makeBackground:qc->img method:qc->opt.background];
		return NO;
	}
	sz = [qc->img size];
	if (qc->opt.overlap == 0 || !forward)
		[backView drawBackground];
	if (qc->opt.texttransp == 0)
		[self drawRectArea:sz kind:qc->kind];
	[backView drawSover:qc->img];
	if (qc->opt.title)
		[self showTitle:qc imageSize:sz];
	if (qc->opt.mark)
		[backView drawMark:(qc->opt.mark & 0x0f)];
	[backWin display];
	return YES;
}

/* Local Method */
- (void)fetchImage:(struct qcell *)qc from:(NSString *)fname
{
	NSString *fn;
	NSSize	sz;	/* size of the image */
	float	rw, rh;
	BOOL	sizeFixed = NO;
	int	mul;
	double	ratio;

	qc->kind = img_Pict;
	fn = [ohpPath stringByAppendingPathComponent:fname];
	if ([TextImageCreator kindOfFile:fn] == TxImage) {
		if ([[fn pathExtension]
			caseInsensitiveCompare: @"eps"] == NSOrderedSame)
			qc->kind = img_EPS;
		qc->img = [[NSImage alloc] initWithContentsOfFile:fn];
		[qc->img setScalesWhenResized:YES];
		if ((mul = qc->opt.imageSz) > 0) {	/* Image Size fixed */
			sz = [qc->img size];
			if (mul <= im_Integer) { /* Integer & 1-4 */
				if (mul == im_Integer) {
					ratio = [backView enlargeScale:sz
						by:bk_FitScreen err:NULL];
					mul = (ratio < 1.0) ? 1 : (int)ratio;
				}
				if (mul > 1) {
					sz.width *= mul;
					sz.height *= mul;
					[qc->img setSize: sz];
				}
			}else { /* Screen */
				int k = (mul == im_Screen)
					? bk_FitScreen : bk_CoverScreen;
				ratio = [backView enlargeScale:sz
						by:k err:NULL];
				if (k == bk_FitScreen || ratio > 1.0) {
					sz.width = (int)(sz.width * ratio);
					sz.height = (int)(sz.height * ratio);
					[qc->img setSize: sz];
				}
			}
			sizeFixed = YES;
		}
	}else {
		id creat = [[TextImageCreator alloc] init];
		if (qc->opt.notextclr == 0)
			[creat setTextColor: setting->txcolor];
		[creat setRaiseBaseline:(qc->opt.baseline ?
				qc->opt.baseline : setting->baseline)];
		[creat setFont:[setting->txfont fontName]
			size: (qc->opt.fontsize ?
				qc->opt.fontsize : setting->fontsize)];
		qc->img = [creat textImageFrom:fn];
		[qc->img retain];
		[qc->img setScalesWhenResized:YES];
		[creat release];
		qc->kind = img_Text;

	 	/* Resize: Only FitScreen is allowed */
		if ((mul = qc->opt.imageSz) == im_Screen) {
			sz = [qc->img size];
			ratio = [backView enlargeScale:sz
						by:bk_FitScreen err:NULL];
			sz.width = (int)(sz.width * ratio);
			sz.height = (int)(sz.height * ratio);
			[qc->img setSize: sz];
			sizeFixed = YES;
		}
	}
	if (qc->opt.backgflag) /* Background Image */
		return;

	if (!sizeFixed) {
		float bodyRatio = [[PrefCtrl defaultPref] bodyRatio];
		sz = [qc->img size];
		rw = scrRect.size.width * bodyRatio / sz.width;
		rh = scrRect.size.height * bodyRatio / sz.height;
		if (rw > rh) rw = rh;
		[qc->img setScalesWhenResized:YES];
		sz.width = (int)(sz.width * rw);
		sz.height = (int)(sz.height * rw);
		[qc->img setSize:sz];
		sizeFixed = YES;
	}
}

/* Local Method */
- (BOOL)getQEntry:(int)index
{
	/* Note that this method is not guarded by LOCK */
	struct qcell	*q;
	NSString *fn;

	if (qarray[index] == NULL) {
		qarray[index] = q = [self allocQcell];
		fn = [self fetchFileAndOpt:&q->opt atRow:index];
		if (fn != nil) {
			[self fetchImage:q from:fn];
			if (q->opt.title)
				[self setTitleAtIndex:index];
			else
				q->title = nil;
			q->skip = NO;
		}else
			q->skip = YES;
	}
	return !qarray[index]->skip;
}

/* Local Method */
- (void)readAhead:(id)sender
{
	int ahead, cnt;
	BOOL	sw;
	NSAutoreleasePool *pool;

	pool =[[NSAutoreleasePool alloc] init];
	cnt = 0;
	for (ahead = currentIndex + 1; ahead < sheetNum; ahead++) {
		[qLock lock];
		if ((sw = [self getQEntry:ahead]) == YES)
			if (qarray[ahead]->opt.backgflag)
				sw = NO;
		[qLock unlock];
		if (sw && ++cnt >= QForward)
			break;
	}
	[pool release];
}

- (void)startFrom:(int)index
{
	sheetNum = [table numberOfRows];
	qarray = (struct qcell **)calloc(sheetNum, sizeof(struct qcell *));
	currentIndex = index - 1;
	[backView drawBackground];
	[backWin makeKeyAndOrderFront:self];
	[backWin display];
	[self nextPush:self];
}

- (void)nextPush:(id)sender
{
	int	del;
	BOOL	sw;

	do {
	    for (++currentIndex ; currentIndex < sheetNum; currentIndex++) {
		del = currentIndex - QBackward - 1;
		if (del >= 0) {
			[qLock lock];
			if (qarray[del] != NULL) {
				[self freeQcell: qarray[del]];
				qarray[del] = NULL;
			}
			[qLock unlock];
		}
		[qLock lock];
		sw = [self getQEntry:currentIndex];
		[qLock unlock];
		if (sw)
			break;
	    }
	    if (currentIndex >= sheetNum) {
		currentIndex = sheetNum - 1;
		/* could go on with "prevPush" */
		/* Note: Don't call prevPush: here to prevent loop. */
		NSBeep();
		return;	/* End of Sheets */
	    }
	    [table selectRow:currentIndex byExtendingSelection:NO];

	}while (![self showSheet:qarray[currentIndex] forward:YES]);
#ifdef Thread_Safe
	/* Image Filtering Service of NSImage is not thread-safe.
	   It uses NSConnection in the main thread.  Other threads can't
	   access this connection.
	*/
	if (currentIndex < sheetNum - 1)
		[NSThread detachNewThreadSelector:@selector(readAhead:)
			toTarget:self withObject:self];
#else
	if (currentIndex < sheetNum - 1) {
		[NSObject cancelPreviousPerformRequestsWithTarget:self
			selector:@selector(readAhead:) object:self];
		[self performSelector:@selector(readAhead:)
			withObject:self afterDelay:0.2];
	}
#endif
}

- (void)prevPush:(id)sender
{
	int	del;
	BOOL	sw;

	do {
	    for (--currentIndex ; currentIndex >= 0; currentIndex--) {
		del = currentIndex + QForwardMax + 1;
		if (del < sheetNum) {
			[qLock lock];
			if (qarray[del] != NULL) {
				[self freeQcell: qarray[del]];
				qarray[del] = NULL;
			}
			[qLock unlock];
		}
		[qLock lock];
		if ((sw = [self getQEntry:currentIndex]) == YES)
			if (qarray[currentIndex]->opt.backgflag)
				sw = NO;
		[qLock unlock];
		if (sw)
			break;
	    }
	    if (currentIndex < 0) {
		currentIndex = -1;	/* can go on with "nextPush" */
		[self nextPush:self];
		NSBeep();
		return;	/* Beginning of Sheets */
	    }
	    [table selectRow:currentIndex byExtendingSelection:NO];
	}while (![self showSheet:qarray[currentIndex] forward:NO]);
}

- (void)quitPush:(id)sender
{
	[backWin close];	/* backWin is release'd */
	backWin = nil;
	[self freeQarray];
	[quitTarget performSelector:quitSel];
	/* default action is [self release] */
}


@end
