#import "TextImageCreator.h"
#import <Foundation/NSString.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSColor.h>
#import <AppKit/psops.h>
#import <AppKit/NSTextView.h>
#import <AppKit/NSLayoutManager.h>
#import <AppKit/NSWindow.h>
#import <AppKit/NSFont.h>
#import "SettingDictionary.h"
#import "PrefCtrl.h"
#import <stdio.h>


/* These parameters are used to draw text into internal view. */
static NSSize normalTextViewSize = { 480.0, 360.0 };
static float ratioLimit = 1.0;
static NSString *wraptypes[] = { @"rtfd", nil };
static NSString *bintypes[] = {
	@"a",	@"app",	@"compressed",	@"dvi",	@"gz",
	@"nib",	@"o",	@"pkg",	@"tar",	@"Z", nil
};

@implementation TextImageCreator

+ (int)kindOfFile:(NSString *)path
{
	NSString *ex = [path pathExtension];
	if ([ex caseInsensitiveCompare: @"rtfd"] == NSOrderedSame)
		return TxRTFD;
	if ([ex caseInsensitiveCompare: @"rtf"] == NSOrderedSame)
		return TxRTF;
	if ([[NSImage imageFileTypes] containsObject:ex])
		return TxImage;
	return TxPlane;
}

+ (BOOL)isWrappedFile:(NSString *)fname
{
	int	i;
	NSString *ex = [fname pathExtension];
	for (i = 0; wraptypes[i]; i++)
		if ([ex isEqualToString:wraptypes[i]])
			return YES;
	return NO;
}

+ (BOOL)isBinaryFile:(NSString *)fname
{
	int	i;
	NSString *ex = [fname pathExtension];
	for (i = 0; bintypes[i]; i++)
		if ([ex isEqualToString:bintypes[i]])
			return YES;
	return NO;
}

#define  KMAX	12

+ (NSSize)sizeOfRTF:(NSString *)fname RTFD:(BOOL)isrtfd
{
	NSSize	sz;
	int	cc, i, v;
	FILE	*fp;
	char	buf[KMAX];

	sz.width = sz.height = 0;
	if (isrtfd)
		fname = [fname stringByAppendingPathComponent:@"TXT.rtf"];
	if ((fp = fopen([fname cString], "r")) == NULL)
		return sz;	/* ERROR */
	cc = getc(fp);
	while (cc != EOF) {
		if (cc != '\\') {
			cc = getc(fp);
			continue;
		}
		cc = getc(fp);
		for (i = 0; i < KMAX-1 && cc >= 'a' && cc <= 'z'; i++) {
			buf[i] = cc;
			cc = getc(fp);
		}
		buf[i] = 0;
		for (v = 0; cc >= '0' && cc <= '9'; ) {
			v = (v * 10) + cc - '0';
			cc = getc(fp);
		}
		if (strcmp(buf, "paperw") == 0)
			sz.width = v / 20;
		else if (strcmp(buf, "paperh") == 0)
			sz.height = v / 20;
		if (sz.width > 0 && sz.height > 0)
			break;
	}
	fclose(fp);
	return sz;
}

+ (void)setRatioLimit:(float)ratio
{
	ratioLimit = ratio;
}

+ (void)setNormalTextViewSize:(NSSize)size
{
	normalTextViewSize = size;
}


- (id)init
{
	struct setinfo	*info;

	[super init];
	info = [[PrefCtrl defaultPref] defaultSetting];
	baseline = info->baseline;
	fontsize = realFontSize(info->fontsize);
	currentfont = info->txfont;
	textColor = nil;
	return self;
}

- (void)dealloc
{
	[textColor release];
	[super dealloc];
}

- (void)setTextColor:(NSColor *)color
{
	[textColor release];
	textColor = color;
	[textColor retain];
}

- (void)setRaiseBaseline:(int)val
{
	baseline = val;
}

- (void)setFont:(NSString *)fontname size:(int)size
{
	fontsize = realFontSize(size);
	currentfont = [NSFont fontWithName:fontname size:fontsize];
}

- (NSImage *)textImageFrom:(NSString *)textfile
{
	NSTextView *textView;
	NSRect	rect;
	NSData *dat;
	NSWindow *win;
	NSImage *img;
	int	kind;

	if (textfile == nil)
		return nil;
	if ((kind = [[self class] kindOfFile:textfile]) == TxImage)
		return nil;
		
	rect.origin.x = rect.origin.y = 0;
	if (kind == TxRTF || kind == TxRTFD) {
		NSSize	rtfsize;
		rtfsize = [[self class] sizeOfRTF:textfile RTFD:(kind == TxRTFD)];
		if (rtfsize.height <= 0 || rtfsize.width <= 0)
			return nil;	/* Error */
		if (rtfsize.height / rtfsize.width > ratioLimit)
			rtfsize.height = (int)(rtfsize.width * ratioLimit);
		rect.size = rtfsize;
	}else {
		rect.size = normalTextViewSize;
	}
	textView = [[NSTextView alloc] initWithFrame: rect];
	[textView autorelease];
	[textView setDrawsBackground: NO];
	if (kind == TxPlane) {
		NSString *tx;
		NSRange	rng;
		int	i;

		rng = NSMakeRange(0, 0 /* [[textView string] length] */);
		if ((tx = [NSString stringWithContentsOfFile:textfile]) == nil)
			return nil;
		[textView replaceCharactersInRange:rng withString:tx];
		[textView selectAll:self];
		[textView setFont:currentfont];
		if (textColor)
			[textView setTextColor:textColor];
	/*
		if (baseline == 1) {
			for (i = 0; i < 5; i++)
				[textView raiseBaseline:self];
		}else if (baseline > 1) {
			for (i = 1; i < baseline; i++)
				[textView superscript:self];
		}
	*/
		if (baseline > 0) {
			static unsigned char rep[] = {
				5, 7, 9, 11, 13, 16, 19, 22};
			for (i = rep[baseline-1]; i > 0; i--)
				[textView raiseBaseline:self];
		}
	}else { /* RTF, RTFD */
		[textView readRTFDFromFile:textfile];
		if (textColor)
			[textView setTextColor:textColor];
		// [textView sizeToFit];
		rect = [textView bounds];
	}
	win = [[NSWindow alloc] initWithContentRect:rect styleMask:0
		backing:NSBackingStoreRetained defer:NO];
	[win setContentView:textView];
	[win display];
	// "display" is needed to show the text at correct point.
	dat = [textView dataWithEPSInsideRect:rect];
	img = [[[NSImage alloc] initWithData:dat] autorelease];
	[win release];
	return img;
}

- (NSImage *)textImageFromString:(NSString *)string
{
	NSTextView *textView;
	NSData *dat;
	NSWindow *win;
	NSImage *img;
	NSRect	rect;
	NSSize	scsz, sz;
	float	ratio;

	if (string == nil || [string length] == 0)
		return nil;
		
	scsz = [[PrefCtrl defaultPref] screenSize];
	rect.origin.x = rect.origin.y = 0;
	/* initial rect size */
	rect.size.width = scsz.width;
	rect.size.height = fontsize * 2;
	textView = [[NSTextView alloc] initWithFrame: rect];
	[textView setMinSize:rect.size];
	[textView setMaxSize:(NSSize){2000, 1e4}];
	[textView setVerticallyResizable:NO];
	[textView setHorizontallyResizable:YES];
	// [textView setAutoresizingMask:NSViewWidthSizable];
	// [[textView textContainer] setWidthTracksTextView:NO];
	// [[textView textContainer] setHeightTracksTextView:NO];
	[textView autorelease];
	[textView setDrawsBackground: NO];
	[textView setString:string];
	[textView selectAll:self];
	[textView setFont:currentfont];
	if (textColor)
		[textView setTextColor:textColor];
	[textView sizeToFit];
	rect = [[textView layoutManager]
		usedRectForTextContainer:[textView textContainer]];
	// rect = [textView frame];
	win = [[NSWindow alloc] initWithContentRect:rect styleMask:0
		backing:NSBackingStoreRetained defer:NO];
	[win setContentView:textView];
	[win display];
	// "display" is needed to show the text at correct point.
	dat = [textView dataWithEPSInsideRect:rect];
	img = [[[NSImage alloc] initWithData:dat] autorelease];
	[win release];
	[img setScalesWhenResized:YES];
	ratio = [[PrefCtrl defaultPref] bodyRatio] * scsz.width
			/ normalTextViewSize.width;
	sz.width = rect.size.width * ratio;
	sz.height = rect.size.height * ratio;
	[img setSize:sz];
	return img;
}

@end

